@@ -3,7 +3,7 @@
  *
  * SPDX-License-Identifier: LGPL-2.1-or-later
  *
- * Author: Alexander Mikhaylenko <alexander.mikhaylenko@puri.sm>
+ * Author: Alice Mikhaylenko <alice.mikhaylenko@puri.sm>
  */
 
 #include "config.h"
@@ -144,12 +144,10 @@ update_indicator (AdwTabThumbnail *self)
   set_style_class (GTK_WIDGET (self), "indicator", indicator != NULL);
 }
 
-static gboolean
+static void
 close_idle_cb (AdwTabThumbnail *self)
 {
   adw_tab_view_close_page (self->view, self->page);
-
-  return G_SOURCE_REMOVE;
 }
 
 static void
@@ -161,15 +159,13 @@ close_clicked_cb (AdwTabThumbnail *self)
   /* When animations are disabled, we don't want to immediately remove the
    * whole tab mid-click. Instead, defer it until the click has happened.
    */
-  g_idle_add ((GSourceFunc) close_idle_cb, self);
+  g_idle_add_once ((GSourceOnceFunc) close_idle_cb, self);
 }
 
-static gboolean
+static void
 unpin_idle_cb (AdwTabThumbnail *self)
 {
   adw_tab_view_set_page_pinned (self->view, self->page, FALSE);
-
-  return G_SOURCE_REMOVE;
 }
 
 static void
@@ -181,7 +177,7 @@ unpin_clicked_cb (AdwTabThumbnail *self)
   /* When animations are disabled, we don't want to immediately unpin the
    * whole tab mid-click. Instead, defer it until the click has happened.
    */
-  g_idle_add ((GSourceFunc) unpin_idle_cb, self);
+  g_idle_add_once ((GSourceOnceFunc) unpin_idle_cb, self);
 }
 
 static void
@@ -213,8 +209,10 @@ drop_cb (AdwTabThumbnail *self,
          GValue          *value)
 {
   gboolean ret = GDK_EVENT_PROPAGATE;
+  GdkDrop *drop = gtk_drop_target_get_current_drop (self->drop_target);
+  GdkDragAction preferred_action = gdk_drop_get_actions (drop);
 
-  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, &ret);
+  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, preferred_action, &ret);
 
   return ret;
 }
@@ -523,8 +521,9 @@ adw_tab_thumbnail_class_init (AdwTabThumbnailClass *klass)
                   g_signal_accumulator_first_wins,
                   NULL, NULL,
                   G_TYPE_BOOLEAN,
-                  1,
-                  G_TYPE_VALUE);
+                  2,
+                  G_TYPE_VALUE,
+                  GDK_TYPE_DRAG_ACTION);
 
   signals[SIGNAL_EXTRA_DRAG_VALUE] =
     g_signal_new ("extra-drag-value",
